"""Module with classes that represent the API data"""

from warnings import warn
from datetime import datetime
import pytz

from .types.time_formats import F1, F2
from .errors import (InvalidStrIndexError, InvalidIndexTypeError,
                     InvalidDatetimeIndexError, EmptyInstanceError)


class BaseData:
    """
    Abstract class that represents data in any section of the response

    Attributes
    ----------
    _timezone : str
        String identifier of used timezone

    Methods
    -------
    load_data
        Load the API data from dict to member variables
    get_members
        Get names of member variables that are available
    to_pandas
        Export the data to pandas DataFrame
    """
    def __init__(self, timezone):
        """
        :param str: String identifier of used timezone
        """
        # Prevent users from instantiating the BaseData
        if type(self) is BaseData:  # pylint: disable=C0123
            raise Exception('BaseData cannot be instantiated!')
        # Save the timezone of the data
        self._timezone = timezone

    def load_data(self, data):
        """
        Load the API section data from dict to member variables

        All the members of the 'data' dicts that are not dicts (the direct)
        values are stored into member variables. The nested data (e.g.
        wind.angle) are stored as SingleTimeData instances.

        All dates are converted from strings to 'datetime's. Those dates that
        contain times are localized to be timezone-aware.

        :param dict: The dictionary as retrieved from RequestHandler
        """
        # Keep empty instance if no data are available
        if data is None:
            return

        # Initialize pytz timezone object using the data's timezone
        tz = pytz.timezone(self._timezone)

        # Iterate over the keys of the 'data' dict
        for key, value in data.items():
            # For nested data, initialize SingleTimeData instance
            if isinstance(value, dict):
                # Save the data into the instance attribute
                setattr(self, key, SingleTimeData(value, self._timezone))
            else:
                # None values can occur in the data
                if value is not None:
                    # Items that contain datetime and need to be localized
                    if key in ('date', 'rise', 'set'):
                        # Convert to datetime
                        dt = datetime.strptime(value, F1)
                        # Localize from UTC
                        value = pytz.utc.localize(dt).astimezone(tz)
                    # Items that only contain day (not hours, etc.)
                    if key == 'day':
                        # Only convert to datetime
                        value = datetime.strptime(value, F2)

                # Save the data into the instance attribute
                setattr(self, key, value)

    def get_members(self):
        """
        Build list of variable names that are available in the instance

        :return list: List with string names of the available variables
        """
        # Ignore callables and variables that start with '_'
        return [x for x in dir(self) if
                not callable(getattr(self, x)) and not x.startswith("_")]

    def to_pandas(self):
        """
        Export the data to pandas DataFrame

        NOTE: This needs 'pandas' module, which is not needed for any other
        parts of pymeteosource. Because of this, it is not installed by default
        setup.py to keep the dependencies as minimal as possible. To use this
        feature, use 'pip install pymeteosource[pandas]' to install this
        package, or install pandas manually using 'pip install pandas'.

        :return pandas.DataFrame: The DataFrame with 'date'/'day' as index
        """
        # Try to lazy-load 'pandas' module
        try:
            import pandas as pd  # pylint: disable=C0415
        except ImportError:
            # If pandas is not installed, print warning and return None
            warn("Module pandas is not installed, cannot export the data. "
                 "Try to install pandas with 'pip install pandas'.")
            return None

        # This is the basic use case - rows are the timesteps
        if isinstance(self, MultipleTimesData):
            # pylint: disable=E1101
            df = pd.DataFrame([x.to_dict() for x in self.data])
        # This is also possible, but results in 1-row DataFrame
        elif isinstance(self, SingleTimeData):
            df = pd.DataFrame([self.to_dict()])  # pylint: disable=E1101
        # This should not happen
        else:
            c = self.__class__.__name__
            raise NotImplementedError("Cannot export {} to pandas.".format(c))

        # If 'date' or 'day' column is present, set it as index
        if 'date' in df.columns:
            df.set_index('date', inplace=True)
        elif 'day' in df.columns:
            df.set_index('day', inplace=True)

        return df

    def __repr__(self):
        """
        Override __repr__ to have usefull text when attepting to print
        """
        c, members = self.__class__.__name__, self.get_members()
        date = getattr(self, 'date', getattr(self, 'day', 'current'))
        return ('<Instance of {} ({}) with {} member variables ({})>'.format(
            c, date, len(members), ', '.join(members)))

    def __getitem__(self, attr):
        """
        Override __getitem__ to allow variable access using [] operator
        """
        return getattr(self, attr)


class SingleTimeData(BaseData):
    """
    Class that represents data in single point in time

    This class is used to represent 'current' section, and the individual
    timesteps of 'minutely', 'hourly' and 'daily'.

    Methods
    -------
    to_dict
        Export the data to flat (not nested) dict
    """
    def __init__(self, data, timezone):
        # Call the parent's constructor to initialize the timezone
        super().__init__(timezone)
        # Load the data
        self.load_data(data)

    def to_dict(self, prefix=''):
        """
        Export the data to flat (not nested) dict

        To flatten the nested structures, the variable names are concatenated
        with '_'. For example, wind['angle'] becomes 'wind_angle'. In daily
        data, there are more than two levels of the nested data, so for example
        all_day['wind']['angle'] becomes all_day_wind_angle.

        :return dict: The data as dict
        """
        # Dict to hold the result
        res = {}
        # Iterate over the member variables representing the elements
        for key in self.get_members():
            # Get the value of current key (from the member variable)
            val = getattr(self, key)
            # If the data is nested, we recursively expand it
            if isinstance(val, BaseData):
                # Chain the variable name separated by '_' in each level
                res.update(val.to_dict(prefix='{}{}_'.format(prefix, key)))
            else:
                # If we have not-nested type, we simply add it to the dict
                res['{}{}'.format(prefix, key)] = val

        return res


class MultipleTimesData(BaseData):
    """
    Class that represents data in single point in time

    This class is used to represent 'current' section, and the individual
    timesteps of 'minutely', 'hourly' and 'daily'.

    Attributes
    ----------
    summary : str
        Summary of precipitation minutecast - only present for minutely data
    data_type : str
        The type of section the instance represents, e.g. 'minutely'
    data : list
        List of SingleTimeData instances, each containing single timestep
    dates_str : list
        List of 'str' dates of the timesteps (in the same order)
    dates_dt : list
        List of 'datetime' dates of the timesteps (in the same order)
    """
    def __init__(self, data, data_type, timezone):
        # Call the parent's constructor to initialize the timezone
        super().__init__(timezone)
        # Keep empty instance if no data are available
        if data is None:
            return

        # Get the datetime column based on the 'data_type'
        self.data_type = data_type
        if self.data_type == 'daily':
            date_col, form = 'day', F2
        else:
            date_col, form = 'date', F1

        # Build the list of SingleTimeData instances from the data
        self.data = [SingleTimeData(x, self._timezone) for x in data['data']]
        # Build the list of corresponding 'str' dates
        self.dates_str = [x[date_col].strftime(form) for x in self.data]
        # Build the list of corresponding 'datetime' dates
        self.dates_dt = [x[date_col] for x in self.data]

        # If summary is present, save it
        if 'summary' in data:
            self.summary = data['summary']

    def __repr__(self):
        """
        Override __repr__ to have usefull text when attepting to print
        """
        # Shortcut for the class name
        c = self.__class__.__name__
        # For empty instances, we cannot print any info__
        if len(self) == 0:
            return '<Empty instance of {}>'.format(c)
        # For non-empty instances, we print the timesteps range
        return '<Instance of {} ({}) with {} timesteps from {} to {}>'.format(
            c, self.data_type, len(self), self.dates_str[0],
            self.dates_str[-1])

    def __len__(self):
        """
        Override __len__ to support len() calls to the instance
        """
        try:
            return len(self.data)
        except AttributeError:
            return 0

    def __getitem__(self, attr):
        """
        Override __getitem__ to allow variable access using [] operator

        We want to use the operator to acces wanted timestep. We support
        integer (classic 0-based index), string (using YYYY-MM-DDTHH:MM:SS or
        YYYY-MM-DD format) or datetime (both localized or unlocalized).
        """
        if len(self) == 0:
            raise EmptyInstanceError()
        # For ints, we simply return the data with given index
        if isinstance(attr, int):
            return self.data[attr]
        """
        For string, we use 'dates_str' list and return the data with
        'date'/'day' corresponding to first occurence in 'date_str'. On long
        day, there can be two timesteps with the same string representation of
        the datetime, which is why the 'first occurence' is mentioned here.
        """
        if isinstance(attr, str):
            if attr not in self.dates_str:
                raise InvalidStrIndexError(attr)
            return self.data[self.dates_str.index(attr)]
        # For datetimes, we localize it if necessary and use 'dates_dt' list
        if isinstance(attr, datetime):
            # If the datetime is naive
            if attr.tzinfo is None:
                attr = pytz.timezone(self._timezone).localize(attr)
            else:
                # If it is tz-aware, we convert it to the same timezone
                attr = attr.astimezone(pytz.timezone(self._timezone))
            if attr not in self.dates_dt:
                raise InvalidDatetimeIndexError(attr)
            return self.data[self.dates_dt.index(attr)]

        raise InvalidIndexTypeError(attr)


class Forecast:
    """
    Class that represents point forecast


    Attributes
    ----------
    lat : float
        Actual latitude of the point
    lon : float
        Actual longitude of the point
    elevation : int
        Elevation of the location
    timezone : str
        Timezone str identifier in pytz notation
    units : str
        Units set of the data
    current : SingleTimeData
        The data from 'current' section
    minutely : MultipleTimesData
        The data from 'minutely' section
    hourly : MultipleTimesData
        The data from 'hourly' section
    daily : MultipleTimesData
        The data from 'daily' section
    """
    def __init__(self, data, tz):
        lat, lon = data['lat'], data['lon']
        # Parse the lat, lon string values to floats
        self.lat = float(lat[:-1]) if lat[-1] == 'N' else -float(lat[:-1])
        self.lon = float(lon[:-1]) if lon[-1] == 'E' else -float(lon[:-1])
        self.elevation = data['elevation']
        self.timezone = tz
        self.units = data['units']

        self.current = SingleTimeData(data.get('current', None), self.timezone)
        self.minutely = MultipleTimesData(data.get('minutely', None),
                                          'minutely', self.timezone)
        self.hourly = MultipleTimesData(data.get('hourly', None),
                                        'hourly', self.timezone)
        self.daily = MultipleTimesData(data.get('daily', None),
                                       'daily', self.timezone)

    def __repr__(self):
        """
        Override __repr__ to have usefull text when attepting to print
        """
        return '<Forecast for lat: {}, lon: {}>'.format(self.lat, self.lon)

    def __getitem__(self, attr):
        """
        Override __getitem__ to allow variable access using [] operator
        """
        return getattr(self, attr)



"""Module that handles sending the requests to API"""

import requests

from .errors import InvalidRequestError

class RequestHandler:
    """
    Object that handles sending the requests to API

    Attributes
    ----------
    session : requests.sessions.Session
        A session object to send the requests. This can speed-up multiple
        requests within a program run.

    Methods
    -------
    execute_request
        Execute request and return the JSON response
    """

    def __init__(self, key):
        """
        :param str: The API key
        """
        # Initialize the session
        self.session = requests.Session()
        # Automatically add key header to all requests made within the session
        self.session.headers.update({'X-API-Key': key})

    def execute_request(self, url, **params):
        """
        Make a request and return the JSON response

        :param str: URL of the requests (without the parameters)
        :param kwargs: Arguments of the request (lat, lon, ...)
        """
        response = self.session.get(url, params=params)
        if response.status_code != 200:
            raise InvalidRequestError(response)

        data = response.json()

        # We always get the source data in UTC and then convert to local tz
        assert data['timezone'] == 'UTC'

        return data



"""Tests for pymeteosource"""

import os
import sys
from datetime import datetime
from unittest.mock import MagicMock
from os.path import realpath, join, dirname
import pytz
import pytest
import pandas

from pymeteosource.api import Meteosource
from pymeteosource.types import tiers, endpoints, units, sections
from pymeteosource.data import Forecast, SingleTimeData, MultipleTimesData
from pymeteosource.types.time_formats import F1
from pymeteosource.errors import (InvalidArgumentError, InvalidIndexTypeError,
                                  InvalidStrIndexError, EmptyInstanceError,
                                  InvalidDatetimeIndexError)

from .sample_data import SAMPLE_DATA
from .dst_changes_data import LONG_DAY
from .variables_list import (CURRENT, PRECIPITATION_CURRENT, WIND, MINUTELY,
                             HOURLY, CLOUD, PRECIPITATION, PROBABILITY, DAILY,
                             ALL_DAY, PART_DAY, ASTRO, SUN, MOON, STATS,
                             STATS_TEMP, STATS_WIND, STATS_PREC)

sys.path.insert(0, realpath(join(dirname(__file__), "..")))

# Load API key from environment variable
API_KEY = os.environ.get('METEOSOURCE_API_KEY')
if API_KEY is None:
    raise ValueError("You need to provide API key as environment variable.")


def test_to_dst_changes():
    """Test exporting to pandas"""
    m = Meteosource(API_KEY, tiers.PREMIUM)
    # We mock the API requests with sample data
    m.req_handler.execute_request = MagicMock(return_value=LONG_DAY)
    # Get the mocked forecast
    f = m.get_point_forecast(place_id='london', tz='Europe/Prague')
    # Check the ambiguous date is handled properly
    dts = ['2021-10-31T01:00:00', '2021-10-31T02:00:00',
           '2021-10-31T02:00:00', '2021-10-31T03:00:00']
    assert f.hourly.dates_str == dts

    # Check the astro datetimes are OK when DST changes
    tz = pytz.timezone('Europe/Prague')
    dt = tz.localize(datetime(2021, 10, 31, 1, 24, 35), is_dst=None)
    assert f.daily[0].astro.sun.rise == dt
    dt = tz.localize(datetime(2021, 10, 31, 2, 24, 35), is_dst=True)
    assert f.daily[0].astro.sun.set == dt
    dt = tz.localize(datetime(2021, 10, 31, 2, 24, 35), is_dst=False)
    assert f.daily[0].astro.moon.rise == dt
    dt = tz.localize(datetime(2021, 10, 31, 3, 24, 35), is_dst=None)
    assert f.daily[0].astro.moon.set == dt


def test_build_url():
    """Test URL building"""
    url = 'https://www.meteosource.com/api/v1/TIER/point'
    for tier in [tiers.PREMIUM, tiers.STANDARD, tiers.STARTUP, tiers.FREE]:
        m = Meteosource(API_KEY, tier)
        assert m._build_url(endpoints.POINT) == url.replace('TIER', tier)


def test_get_point_forecast_exceptions():
    """Test detection of invalid point specification detection"""
    m = Meteosource(API_KEY, tiers.PREMIUM)
    # We mock the API requests with sample data
    m.req_handler.execute_request = MagicMock(return_value=SAMPLE_DATA)

    # Test invalid place definitions
    with pytest.raises(InvalidArgumentError) as e:
        m.get_point_forecast(place_id='london', lat=50)
    assert str(e.value) == 'Only place_id or lat+lon can be specified!'
    with pytest.raises(InvalidArgumentError) as e:
        m.get_point_forecast(place_id='london', lon=14)
    assert str(e.value) == 'Only place_id or lat+lon can be specified!'
    with pytest.raises(InvalidArgumentError) as e:
        m.get_point_forecast(place_id='london', lat=50, lon=14)
    assert str(e.value) == 'Only place_id or lat+lon can be specified!'
    with pytest.raises(InvalidArgumentError) as e:
        m.get_point_forecast(lat=50)
    assert str(e.value) == 'Only place_id or lat+lon can be specified!'
    with pytest.raises(InvalidArgumentError) as e:
        m.get_point_forecast(lon=14)
    assert str(e.value) == 'Only place_id or lat+lon can be specified!'
    with pytest.raises(InvalidArgumentError) as e:
        m.get_point_forecast()
    assert str(e.value) == 'Only place_id or lat+lon can be specified!'

    # Test valid place definitions
    m.get_point_forecast(place_id='london')
    m.get_point_forecast(lat=50, lon=14)


def test_forecast_indexing():
    """Test indexing MultipleTimesData with int, string and datetimes"""
    m = Meteosource(API_KEY, tiers.PREMIUM)
    # We mock the API requests with sample data
    m.req_handler.execute_request = MagicMock(return_value=SAMPLE_DATA)
    # Get the mocked forecast
    f = m.get_point_forecast(place_id='london', tz='UTC')

    # Index by int
    assert f.hourly[1].wind.angle == 106

    # Index by too large int
    with pytest.raises(IndexError):
        assert f.hourly[1000]

    # Index by string
    assert f.hourly['2021-09-08T11:00:00'].feels_like == 23.2

    # Index by string with wrong format
    with pytest.raises(InvalidStrIndexError):
        f.hourly['2021-09-08 11:00:00']  # pylint: disable=W0104

    # Index by unsupported type
    with pytest.raises(InvalidIndexTypeError):
        f.hourly[1.1]  # pylint: disable=W0104

    # Index by tz-naive datetime
    dt = datetime.strptime('2021-09-09T00:00:00', F1)
    assert f.hourly[dt].probability.precipitation == 61

    # Index by tz-aware datetime but in different tz
    dt1 = pytz.timezone('UTC').localize(dt)
    assert f.hourly[dt1].probability.precipitation == 61

    # Index by tz-aware datetime
    dt1 = pytz.timezone('Europe/London').localize(dt)
    assert f.hourly[dt1].probability.precipitation == 21

    # Index by tz-aware datetime but with wrong timezone
    dt2 = pytz.timezone('Asia/Kabul').localize(dt)
    with pytest.raises(InvalidDatetimeIndexError) as e:
        f.hourly[dt2]  # pylint: disable=W0104
    assert 'Invalid datetime index' in str(e.value)

    # Check timezone settings
    f = m.get_point_forecast(place_id='london', tz='Europe/London')
    dt = pytz.timezone('Europe/London').localize(datetime(2021, 9, 8, 10))
    assert f.hourly[0].date == dt

    f = m.get_point_forecast(place_id='london', tz='Europe/Prague')
    dt = pytz.timezone('Europe/Prague').localize(datetime(2021, 9, 8, 11))
    assert f.hourly[0].date == dt

    f = m.get_point_forecast(place_id='london', tz='Asia/Kabul')
    dt = pytz.timezone('Asia/Kabul').localize(datetime(2021, 9, 8, 13, 30))
    assert f.hourly[0].date == dt


def test_to_pandas():
    """Test exporting to pandas"""
    m = Meteosource(API_KEY, tiers.PREMIUM)
    # We mock the API requests with sample data
    m.req_handler.execute_request = MagicMock(return_value=SAMPLE_DATA)
    # Get the mocked forecast
    f = m.get_point_forecast(place_id='london')

    df = f.current.to_pandas()
    assert len(df) == 1

    df = f.minutely.to_pandas()
    assert len(df) == 116
    assert isinstance(df.index, pandas.core.indexes.datetimes.DatetimeIndex)

    df = f.hourly.to_pandas()
    assert len(df) == 155
    assert isinstance(df.index, pandas.core.indexes.datetimes.DatetimeIndex)

    df = f.daily.to_pandas()
    assert len(df) == 30
    assert isinstance(df.index, pandas.core.indexes.datetimes.DatetimeIndex)


def test_to_dict():
    """Test exporting to pandas"""
    m = Meteosource(API_KEY, tiers.PREMIUM)
    # We mock the API requests with sample data
    m.req_handler.execute_request = MagicMock(return_value=SAMPLE_DATA)
    # Get the mocked forecast
    f = m.get_point_forecast(place_id='london')

    # Test multilevel dict flattening
    assert 'afternoon_wind_angle' in f.daily[0].to_dict()


def test_forecast_structure():
    """Test structure of the Forecast object on real data"""
    # Initialize the Meteosource object
    m = Meteosource(API_KEY, tiers.PREMIUM)
    # Get real forecast data (not mocked)
    f = m.get_point_forecast(place_id='london', tz='Asia/Kabul',
                             units=units.METRIC, sections=sections.ALL)

    # Check if the header is correct
    assert isinstance(f, Forecast)
    assert f.lat == 51.50853
    assert f.lon == -0.12574
    assert f.elevation == 25
    assert f.timezone == 'Asia/Kabul'
    assert f.units == 'metric'

    # Check types of the sections
    assert isinstance(f.current, SingleTimeData)
    assert isinstance(f.minutely, MultipleTimesData)
    assert isinstance(f.hourly, MultipleTimesData)
    assert isinstance(f.daily, MultipleTimesData)

    # Check current section
    assert set(f.current.get_members()) == CURRENT
    assert set(f.current.wind.get_members()) == WIND  # pylint: disable=E1101
    # pylint: disable=E1101
    assert set(f.current.precipitation.get_members()) == PRECIPITATION_CURRENT

    # Check minutely section
    assert isinstance(f.minutely.summary, str)
    assert isinstance(f.minutely, MultipleTimesData)
    assert len(f.minutely) > 0
    assert set(f.minutely[0].get_members()) == MINUTELY
    assert isinstance(f.minutely[0].date, datetime)

    # Check hourly section
    assert isinstance(f.hourly, MultipleTimesData)
    assert len(f.hourly) > 0
    assert set(f.hourly[0].get_members()) == HOURLY
    assert set(f.hourly[0].wind.get_members()) == WIND
    assert set(f.hourly[0].cloud_cover.get_members()) == CLOUD
    assert set(f.hourly[0].precipitation.get_members()) == PRECIPITATION
    assert set(f.hourly[0].probability.get_members()) == PROBABILITY

    # Check daily section
    assert isinstance(f.daily, MultipleTimesData)
    assert set(f.daily[0].get_members()) == DAILY
    assert isinstance(f.daily[0].day, datetime)
    assert set(f.daily[0].all_day.get_members()) == ALL_DAY
    assert set(f.daily[0].all_day.wind.get_members()) == WIND
    assert set(f.daily[0].morning.get_members()) == PART_DAY
    assert set(f.daily[0].morning.wind.get_members()) == WIND
    assert set(f.daily[0].afternoon.get_members()) == PART_DAY
    assert set(f.daily[0].afternoon.wind.get_members()) == WIND
    assert set(f.daily[0].evening.get_members()) == PART_DAY
    assert set(f.daily[0].evening.wind.get_members()) == WIND
    assert set(f.daily[0].astro.get_members()) == ASTRO
    assert set(f.daily[0].astro.sun.get_members()) == SUN
    assert set(f.daily[0].astro.moon.get_members()) == MOON
    assert set(f.daily[0].statistics.get_members()) == STATS
    assert set(f.daily[0].statistics.temperature.get_members()) == STATS_TEMP
    assert set(f.daily[0].statistics.wind.get_members()) == STATS_WIND
    assert set(f.daily[0].statistics.precipitation.get_members()) == STATS_PREC

    # Check correct exception raising when minutely data are not present
    m = Meteosource(API_KEY, tiers.FREE)
    # Get real forecast data (not mocked)
    f = m.get_point_forecast(place_id='london')
    with pytest.raises(EmptyInstanceError) as e:
        f.minutely[0]  # pylint: disable=W0104
    assert str(e.value) == 'The instance does not contain any data!'



"""Definitions of present variables at given places in the response"""

WIND = {'angle', 'dir', 'gusts', 'speed'}

MINUTELY = {'date', 'precipitation'}

CLOUD = {'total', 'low', 'middle', 'high'}

PROBABILITY = {'precipitation', 'storm', 'freeze'}

PRECIPITATION = {'total', 'type', 'convective', 'rainspot'}

CURRENT = {'cloud_cover', 'dew_point', 'feels_like', 'humidity', 'icon',
           'icon_num', 'irradiance', 'ozone', 'precipitation', 'pressure',
           'summary', 'temperature', 'uv_index', 'visibility', 'wind',
           'wind_chill'}

PRECIPITATION_CURRENT = {'total', 'type'}

HOURLY = {'date', 'weather', 'icon', 'summary', 'temperature', 'feels_like',
          'soil_temperature', 'wind_chill', 'dew_point', 'surface_temperature',
          'wind', 'cloud_cover', 'pressure', 'precipitation', 'probability',
          'cape', 'evaporation', 'irradiance', 'lftx', 'ozone', 'uv_index',
          'humidity', 'snow_depth', 'sunshine_duration', 'visibility'}

DAILY = {'day', 'weather', 'icon', 'summary', 'predictability', 'all_day',
         'morning', 'afternoon', 'evening', 'astro', 'statistics'}

ALL_DAY = {'weather', 'icon', 'temperature', 'temperature_min',
           'temperature_max', 'feels_like', 'feels_like_min', 'feels_like_max',
           'soil_temperature', 'soil_temperature_min', 'soil_temperature_max',
           'wind_chill', 'wind_chill_min', 'wind_chill_max', 'dew_point',
           'dew_point_min', 'dew_point_max', 'surface_temperature',
           'surface_temperature_min', 'surface_temperature_max', 'wind',
           'cloud_cover', 'pressure', 'precipitation', 'probability',
           'evaporation', 'irradiance', 'ozone', 'humidity', 'snow_depth',
           'sunshine_duration', 'visibility'}

PART_DAY = {'weather', 'icon', 'temperature', 'feels_like', 'soil_temperature',
            'wind_chill', 'dew_point', 'surface_temperature', 'wind',
            'cloud_cover', 'pressure', 'precipitation', 'probability',
            'evaporation', 'irradiance', 'ozone', 'humidity', 'snow_depth',
            'sunshine_duration', 'visibility'}
ASTRO = {'sun', 'moon'}

SUN = {'rise', 'set', 'always_up', 'always_down'}

MOON = {'rise', 'set', 'always_up', 'always_down', 'phase'}

STATS = {'temperature', 'wind', 'precipitation'}

STATS_TEMP = {'avg', 'avg_min', 'avg_max', 'record_min', 'record_max'}
STATS_WIND = {'avg_speed', 'avg_angle', 'avg_dir', 'max_speed', 'max_gust'}
STATS_PREC = {'avg', 'probability'}



"""Module that provide the Meteosource interface object"""

from .request_handler import RequestHandler
from .types import langs, sections, units, endpoints
from .errors import InvalidArgumentError
from .data import Forecast


class Meteosource:
    """
    The main object that provides the interface for Meteosource API

    Attributes
    ----------
    req_handler : RequestHandler
        RequestHandler object to be used for the requests
    host : string
        The host URL of the Meteosource API
    tier : string
        The tier the user is using

    Methods
    -------
    _build_url
        Build URL for the request
    get_point_forecast
        Get forecast data for given point
    """
    def __init__(self, api_key, tier, host='https://www.meteosource.com/api'):
        """
        Basic constructor

        :param str: API key
        :param str: Tier the user is using
        :param str: Host URL of the Meteosource API
        """
        # Initialize the request handler with the API key
        self.req_handler = RequestHandler(api_key)
        self.host = host
        self.tier = tier

    def _build_url(self, endpoint):
        """
        Build URL for the request

        :param str: Endpoint for the request
        :return str: The URL of the request without parameters (lat, lon, ...)
        """
        pars = {'host': self.host, 'tier': self.tier, 'endpoint': endpoint}
        url = '{host}/v1/{tier}/{endpoint}'.format(**pars)

        return url

    def get_point_forecast(self, place_id=None, lat=None, lon=None,
                           sections=(sections.CURRENT, sections.HOURLY),
                           tz='UTC', lang=langs.ENGLISH, units=units.AUTO):
        """
        Get forecast data for given point

        :param str: Identifier of the place (place_id)
        :param float: Latitude of the point
        :param float: Longitude of the point
        :param str: Sections to return
        :param str: Timezone for final output. Requests are always made in UTC!
        :param str: Language
        :param str: Units to use
        :return Forecast: Forecast object with the forecast data
        """
        # Build the URL for the request
        url = self._build_url(endpoints.POINT)
        if isinstance(sections, (list, tuple)):
            sections = ','.join(sections)
        # Parameters of the request, the requested tz is always UTC!
        pars = {'language': lang, 'units': units, 'timezone': 'UTC',
                'sections': sections}

        # If place_id is not specified, we use lat+lon
        if place_id is None:
            # Check if both lat+lon are specified
            if lat is None or lon is None:
                raise InvalidArgumentError()
            # Add the lat+lon to the parameters
            pars.update({'lat': lat, 'lon': lon})
        else:
            # Check that only place_id is specified
            if lat is not None or lon is not None:
                raise InvalidArgumentError()
            # Add the place_id to the parameters
            pars.update({'place_id': place_id})

        # Execute the request with the built URL and parameters
        data = self.req_handler.execute_request(url, **pars)

        # Load the result into Forecast object and return it
        return Forecast(data, tz)



"""Sample API output dict for testing"""

LONG_DAY = {
    "lat":"51.50853N",
    "lon":"0.12574W",
    "elevation":25,
    "timezone":"UTC",
    "units":"metric",
    "current":{
        "icon":"sunny",
        "icon_num":2,
        "summary":"Sunny",
        "temperature":19.8,
        "feels_like":18.8,
        "wind_chill":19.2,
        "dew_point":14.8,
        "wind":{
            "speed":2.6,
            "gusts":5.7,
            "angle":102,
            "dir":"ESE"
        },
        "precipitation":{
            "total":0.0,
            "type":"none"
        },
        "cloud_cover":10,
        "irradiance":258.62,
        "ozone":285.61,
        "pressure":1015,
        "uv_index":1.09,
        "humidity":81,
        "visibility":9.12
    },
    "minutely":None,
    "hourly":{
        "data":[
            {
                "date":"2021-10-30T23:00:00",
                "weather":"sunny",
                "icon":2,
                "summary":"Sunny",
                "temperature":19.8,
                "feels_like":18.8,
                "soil_temperature":16.8,
                "wind_chill":19.2,
                "dew_point":11.8,
                "surface_temperature":19.0,
                "wind":{
                    "speed":2.6,
                    "gusts":5.7,
                    "dir":"ESE",
                    "angle":102
                },
                "cloud_cover":{
                    "total":10,
                    "low":0,
                    "middle":0,
                    "high":6
                },
                "pressure":1014,
                "precipitation":{
                    "total":0.0,
                    "type":"none",
                    "convective":0.0,
                    "rainspot":".......\n.......\n.......\n.......\n.......\n.......\n.......\n"
                },
                "probability":{
                    "precipitation":0,
                    "storm":0,
                    "freeze":0
                },
                "cape":1,
                "evaporation":0.49,
                "irradiance":258.62,
                "lftx":6.29,
                "ozone":285.61,
                "uv_index":1.09,
                "humidity":63,
                "snow_depth":0.0,
                "sunshine_duration":3600.0,
                "visibility":24.14
            },
            {
                "date":"2021-10-31T00:00:00",
                "weather":"sunny",
                "icon":2,
                "summary":"Sunny",
                "temperature":22.0,
                "feels_like":21.2,
                "soil_temperature":18.2,
                "wind_chill":22.2,
                "dew_point":11.8,
                "surface_temperature":22.0,
                "wind":{
                    "speed":3.4,
                    "gusts":7.0,
                    "dir":"ESE",
                    "angle":106
                },
                "cloud_cover":{
                    "total":10,
                    "low":0,
                    "middle":0,
                    "high":41
                },
                "pressure":1014,
                "precipitation":{
                    "total":0.0,
                    "type":"none",
                    "convective":0.0,
                    "rainspot":".......\n.......\n.......\n.......\n.......\n.......\n.......\n"
                },
                "probability":{
                    "precipitation":0,
                    "storm":0,
                    "freeze":0
                },
                "cape":2,
                "evaporation":0.68,
                "irradiance":371.34,
                "lftx":5.65,
                "ozone":286.95,
                "uv_index":2.14,
                "humidity":55,
                "snow_depth":0.0,
                "sunshine_duration":3600.0,
                "visibility":24.14
            },
            {
                "date":"2021-10-31T01:00:00",
                "weather":"mostly_sunny",
                "icon":3,
                "summary":"Mostly sunny",
                "temperature":24.0,
                "feels_like":23.2,
                "soil_temperature":19.5,
                "wind_chill":24.5,
                "dew_point":11.5,
                "surface_temperature":24.2,
                "wind":{
                    "speed":4.0,
                    "gusts":7.5,
                    "dir":"ESE",
                    "angle":108
                },
                "cloud_cover":{
                    "total":16,
                    "low":0,
                    "middle":0,
                    "high":52
                },
                "pressure":1014,
                "precipitation":{
                    "total":0.0,
                    "type":"none",
                    "convective":0.0,
                    "rainspot":".......\n.......\n.......\n.......\n.......\n.......\n.......\n"
                },
                "probability":{
                    "precipitation":0,
                    "storm":0,
                    "freeze":0
                },
                "cape":1,
                "evaporation":0.9,
                "irradiance":504.41,
                "lftx":5.5,
                "ozone":288.03,
                "uv_index":3.36,
                "humidity":47,
                "snow_depth":0.0,
                "sunshine_duration":3600.0,
                "visibility":24.14
            },
            {
                "date":"2021-10-31T02:00:00",
                "weather":"mostly_sunny",
                "icon":3,
                "summary":"Mostly sunny",
                "temperature":24.0,
                "feels_like":23.2,
                "soil_temperature":19.5,
                "wind_chill":24.5,
                "dew_point":11.5,
                "surface_temperature":24.2,
                "wind":{
                    "speed":4.0,
                    "gusts":7.5,
                    "dir":"ESE",
                    "angle":108
                },
                "cloud_cover":{
                    "total":16,
                    "low":0,
                    "middle":0,
                    "high":52
                },
                "pressure":1014,
                "precipitation":{
                    "total":0.0,
                    "type":"none",
                    "convective":0.0,
                    "rainspot":".......\n.......\n.......\n.......\n.......\n.......\n.......\n"
                },
                "probability":{
                    "precipitation":0,
                    "storm":0,
                    "freeze":0
                },
                "cape":1,
                "evaporation":0.9,
                "irradiance":504.41,
                "lftx":5.5,
                "ozone":288.03,
                "uv_index":3.36,
                "humidity":47,
                "snow_depth":0.0,
                "sunshine_duration":3600.0,
                "visibility":24.14
            },
            ]
    },
    "daily":{
        "data":[
            {
                "day":"2021-09-08",
                "weather":"partly_sunny",
                "icon":4,
                "summary":"Sunny changing to mostly cloudy by evening. Temperature 18/28 °C.",
                "predictability":1,
                "all_day":{
                    "weather":"partly_sunny",
                    "icon":4,
                    "temperature":23.2,
                    "temperature_min":17.5,
                    "temperature_max":28.0,
                    "feels_like":22.8,
                    "feels_like_min":16.2,
                    "feels_like_max":26.8,
                    "soil_temperature":21.8,
                    "soil_temperature_min":15.5,
                    "soil_temperature_max":26.2,
                    "wind_chill":24.2,
                    "wind_chill_min":16.5,
                    "wind_chill_max":29.8,
                    "dew_point":11.8,
                    "dew_point_min":10.8,
                    "dew_point_max":12.8,
                    "surface_temperature":23.2,
                    "surface_temperature_min":16.0,
                    "surface_temperature_max":29.2,
                    "wind":{
                        "speed":3.7,
                        "gusts":8.8,
                        "dir":"E",
                        "angle":100
                    },
                    "cloud_cover":{
                        "total":34,
                        "low":0,
                        "middle":9,
                        "high":34
                    },
                    "pressure":1012,
                    "precipitation":{
                        "total":0.0,
                        "type":"none"
                    },
                    "probability":{
                        "precipitation":1,
                        "storm":0,
                        "freeze":0
                    },
                    "evaporation":0.77,
                    "irradiance":321.14,
                    "ozone":289.27,
                    "humidity":50,
                    "snow_depth":0.0,
                    "sunshine_duration":46533,
                    "visibility":24.14
                },
                "morning":{
                    "weather":"Mostly sunny",
                    "icon":3,
                    "temperature":21,
                    "feels_like":20,
                    "soil_temperature":18,
                    "wind_chill":21,
                    "dew_point":12,
                    "surface_temperature":20.8,
                    "wind":{
                        "speed":3.1,
                        "gusts":8.3,
                        "dir":"ESE",
                        "angle":102
                    },
                    "cloud_cover":{
                        "total":12,
                        "low":0,
                        "middle":0,
                        "high":29
                    },
                    "pressure":1014,
                    "precipitation":{
                        "total":0.0,
                        "type":"none"
                    },
                    "probability":{
                        "precipitation":0,
                        "storm":0,
                        "freeze":0
                    },
                    "evaporation":0.61,
                    "irradiance":311.77,
                    "ozone":286.35,
                    "humidity":59,
                    "snow_depth":0.0,
                    "sunshine_duration":19800,
                    "visibility":24.14
                },
                "afternoon":{
                    "weather":"Mostly sunny",
                    "icon":3,
                    "temperature":27.5,
                    "feels_like":26.2,
                    "soil_temperature":25.0,
                    "wind_chill":29.2,
                    "dew_point":11.2,
                    "surface_temperature":28.5,
                    "wind":{
                        "speed":4.6,
                        "gusts":8.8,
                        "dir":"ESE",
                        "angle":109
                    },
                    "cloud_cover":{
                        "total":15,
                        "low":0,
                        "middle":0,
                        "high":22
                    },
                    "pressure":1011,
                    "precipitation":{
                        "total":0.0,
                        "type":"none"
                    },
                    "probability":{
                        "precipitation":0,
                        "storm":0,
                        "freeze":0
                    },
                    "evaporation":1.2,
                    "irradiance":583.2,
                    "ozone":289.54,
                    "humidity":35,
                    "snow_depth":0.0,
                    "sunshine_duration":18000,
                    "visibility":24.13
                },
                "evening":{
                    "weather":"Mostly cloudy",
                    "icon":5,
                    "temperature":21.2,
                    "feels_like":22.0,
                    "soil_temperature":23.5,
                    "wind_chill":23.0,
                    "dew_point":12.2,
                    "surface_temperature":21.2,
                    "wind":{
                        "speed":3.4,
                        "gusts":7.9,
                        "dir":"E",
                        "angle":88
                    },
                    "cloud_cover":{
                        "total":79,
                        "low":0,
                        "middle":28,
                        "high":53
                    },
                    "pressure":1009,
                    "precipitation":{
                        "total":0.0,
                        "type":"none"
                    },
                    "probability":{
                        "precipitation":3,
                        "storm":0,
                        "freeze":0
                    },
                    "evaporation":0.53,
                    "irradiance":70.33,
                    "ozone":292.5,
                    "humidity":55,
                    "snow_depth":0.0,
                    "sunshine_duration":8733,
                    "visibility":24.14
                },
                "astro":{
                    "sun":{
                        "rise":"2021-10-30T23:24:35",
                        "set":"2021-10-31T00:24:35",
                        "always_up":False,
                        "always_down":False
                    },
                    "moon":{
                        "phase":"new_moon",
                        "rise":"2021-10-31T01:24:35",
                        "set":"2021-10-31T02:24:35",
                        "always_up":False,
                        "always_down":False
                    }
                },
                "statistics":{
                    "temperature":{
                        "avg":15.2,
                        "avg_min":11.8,
                        "avg_max":19.2,
                        "record_min":7.0,
                        "record_max":25.8
                    },
                    "wind":{
                        "avg_speed":3.6,
                        "avg_angle":245,
                        "avg_dir":"WSW",
                        "max_speed":7.2,
                        "max_gust":15.2
                    },
                    "precipitation":{
                        "avg":0.0,
                        "probability":24
                    }
                }
            }]
    }
}



from setuptools import find_packages, setup


setup(
    version="0.1",
    name="pymeteosource",
    packages=find_packages(),
    install_requires=["wheel", "requests", "pytz"],
    extras_require={"pandas": "pandas"},
    description="Meteosource API wrapper library",
    author="Meteosource",
    author_email="support@meteosource.com",
    url="https://github.com/Meteosource/pymeteosource/archive/v_01.tar.gz",
    license="MIT",
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""Various custom exceptions"""

class InvalidRequestError(ValueError):
    """
    Exception that is raised when API does not return 200 to a request
    """
    def __init__(self, response):
        code, msg = response.status_code, response.text
        self.message = 'API returned code %s with error: \n  %s' % (code, msg)
        super().__init__(self.message)


class InvalidArgumentError(ValueError):
    """
    Exception that is raised when illegal combination
    of place_id and lat+lon is requested.

    Attributes
    ----------
    MSG : string
        The exceptions message to print
    """
    MSG = 'Only place_id or lat+lon can be specified!'

    def __init__(self):
        self.message = self.MSG
        super().__init__(self.message)


class EmptyInstanceError(ValueError):
    """
    Exception that is raised when access using [] attempted to empty instance

    Attributes
    ----------
    MSG : string
        The exceptions message to print
    """
    MSG = 'The instance does not contain any data!'

    def __init__(self):
        self.message = self.MSG
        super().__init__(self.message)


class InvalidIndexTypeError(ValueError):
    """
    Exception that is raised when illegal index type is used
    for indexing MultipleTimesData.

    Attributes
    ----------
    MSG : string
        The exceptions message to print
    """
    MSG = 'Invalid index data type "%s" to MultipleTimesData!'

    def __init__(self, parameter):
        """
        :param str: The value that was incorrectly used
        """
        self.message = self.MSG % parameter
        super().__init__(self.message)


class InvalidDatetimeIndexError(ValueError):
    """
    Exception that is raised when datetime used to index MultipleTimesData
    is not present in the data.

    Attributes
    ----------
    MSG : string
        The exceptions message to print
    """
    MSG = 'Invalid datetime index "%s" to MultipleTimesData!'

    def __init__(self, parameter):
        """
        :param datetime: The value that was incorrectly used
        """
        self.message = self.MSG % parameter
        super().__init__(self.message)


class InvalidStrIndexError(ValueError):
    """
    Exception that is raised when string used to index MultipleTimesData
    is not present in the data.

    Attributes
    ----------
    MSG : string
        The exceptions message to print
    """
    MSG = 'Invalid string index "%s" to MultipleTimesData!'

    def __init__(self, parameter):
        """
        :param datetime: The value that was incorrectly used
        """
        self.message = self.MSG % parameter
        super().__init__(self.message)
